#include <zombie_escape>

// Definitions 
#define MODEL_MAX_LENGTH 64
#define SOUND_MAX_LENGTH 32

#define GRENADE_ID 4444.0

// Setting File 
new const SETTING_FILE[] = {"zombie_escape.ini"}

// Default Models.
new g_v_szFlareModel[MODEL_MAX_LENGTH] = {"models/v_smokegrenade.mdl"}
new g_p_szFlareModel[MODEL_MAX_LENGTH] = {"models/p_smokegrenade.mdl"}
new g_w_szFlareModel[MODEL_MAX_LENGTH] = {"models/w_smokegrenade.mdl"}

// Default Nade sprite(s).
new g_szExplodeSpr[MODEL_MAX_LENGTH] = {"sprites/fexplo.spr"}
new g_szTrailSpr[MODEL_MAX_LENGTH] = {"sprites/laserbeam.spr"}
new g_szRingSpr[MODEL_MAX_LENGTH] = "sprites/shockwave.spr"

// Default explode sound.
new const szExplodeSound[][SOUND_MAX_LENGTH] = {
	"weapons/explode3.wav",
	"weapons/explode4.wav",
	"weapons/explode5.wav"
}

// Cvars Variables 
new g_pCvarNadeRaduis, 
		g_pCvarNadeDamage,
		g_pCvarHitType,
		g_pCvarExplodeSpr

// Global Variables		
new g_iTrailSpr, 
		g_iExplodeSpr,
		g_iRingSpr

// Dynamic Arrays
new Array:g_szExplodeSound

// Forward called after server activation.
public plugin_init()
{
	// Load plugin 
	register_plugin("[ZE] Grenade: Explode Nade", "1.1", "z0h1r-LK")

	// Fakemeta
	register_forward(FM_SetModel, "SetModel_Pre")

	// Ham 
	RegisterHam(Ham_Think, "grenade", "GrenadeThink_Post", 1)

	// Cvars 
	g_pCvarNadeRaduis = register_cvar("ze_expnade_raduis", "300")
	g_pCvarNadeDamage = register_cvar("ze_expnade_damage", "500")
	g_pCvarExplodeSpr = register_cvar("ze_expnade_explode", "1")
	g_pCvarHitType = register_cvar("ze_expnade_hit_type", "1")
}

// Forward allows precaches game files (models, sounds, generic files).
public plugin_precache() {
	// Load models from external file.
	if (!amx_load_setting_string(SETTING_FILE, "Weapon Models", "V_EXPNADE EXPLODE", g_v_szFlareModel, charsmax(g_v_szFlareModel)))
		amx_save_setting_string(SETTING_FILE, "Weapon Models", "V_EXPNADE EXPLODE", g_v_szFlareModel)
	if (!amx_load_setting_string(SETTING_FILE, "Weapon Models", "P_EXPNADE EXPLODE", g_p_szFlareModel, charsmax(g_p_szFlareModel)))
		amx_save_setting_string(SETTING_FILE, "Weapon Models", "P_EXPNADE EXPLODE", g_p_szFlareModel)
	if (!amx_load_setting_string(SETTING_FILE, "Weapon Models", "W_EXPNADE EXPLODE", g_w_szFlareModel, charsmax(g_w_szFlareModel)))
		amx_save_setting_string(SETTING_FILE, "Weapon Models", "W_EXPNADE EXPLODE", g_w_szFlareModel)

	// Precache models. 
	precache_model(g_v_szFlareModel)
	precache_model(g_p_szFlareModel)
	precache_model(g_w_szFlareModel)

	// Load sprite(s) from external file.
	if (!amx_load_setting_string(SETTING_FILE, "Grenade Sprites", "EXPNADE EXPLODE", g_szExplodeSpr, charsmax(g_szExplodeSpr)))	
		amx_save_setting_string(SETTING_FILE, "Grenade Sprites", "EXPNADE EXPLODE", g_szExplodeSpr)
	if (!amx_load_setting_string(SETTING_FILE, "Grenade Sprites", "TRAIL", g_szTrailSpr, charsmax(g_szTrailSpr)))
		amx_save_setting_string(SETTING_FILE, "Grenade Sprites", "TRAIL", g_szTrailSpr)
	if (!amx_load_setting_string(SETTING_FILE, "Grenade Sprites", "RING", g_szRingSpr, charsmax(g_szRingSpr)))
		amx_save_setting_string(SETTING_FILE, "Grenade Sprites", "RING", g_szRingSpr)

	// Precache sprite(s) and get modelindex.
	g_iExplodeSpr = precache_model(g_szExplodeSpr)
	g_iTrailSpr = precache_model(g_szTrailSpr)
	g_iRingSpr = precache_model(g_szRingSpr)

	// Initialize dynamic arrays.
	g_szExplodeSound = ArrayCreate(SOUND_MAX_LENGTH, 1)

	// Load sound(s) from external file.
	amx_load_setting_string_arr(SETTING_FILE, "Sounds", "EXPNADE EXPLODE", g_szExplodeSound)

	// Save sound(s) in externel file.
	new iIndex
	if (ArraySize(g_szExplodeSound) == 0) {
		for (iIndex = 0; iIndex < sizeof (szExplodeSound); iIndex++)
			ArrayPushString(g_szExplodeSound, szExplodeSound[iIndex])

		// Save sound(s) from external file. 
		amx_save_setting_string_arr(SETTING_FILE, "Sounds", "EXPNADE EXPLODE", g_szExplodeSound)
	}

	// Precache sound. 
	new szSound[SOUND_MAX_LENGTH]
	for (iIndex = 0; iIndex < ArraySize(g_szExplodeSound); iIndex++) {
		// Get string from dynamic array.
		ArrayGetString(g_szExplodeSound, iIndex, szSound, charsmax(szSound))
		
		// Precache sound.
		precache_sound(szSound)
	}
}

// Forward called when set player human.
public ze_user_humanized(id) {
	// Set model(s) for smokegrenade (v_ || p_).
	cs_set_player_view_model(id, CSW_SMOKEGRENADE, g_v_szFlareModel)
	cs_set_player_weap_model(id, CSW_SMOKEGRENADE, g_p_szFlareModel)
}

// Hook called before set entity model.
public SetModel_Pre(iEnt, szModel[]) {
	// We don't care.
	if (strlen(szModel) < 8)
		return FMRES_IGNORED

	// Grenade not thrown yet.
	if (get_entvar(iEnt, var_dmgtime) == 0.0)
		return FMRES_IGNORED

	// Owner of grenade is a zombie.
	if (ze_is_user_zombie(get_entvar(iEnt, var_owner)))
		return FMRES_IGNORED

	// Explode Nade Grenade (Smoke Grenade).
	if (szModel[9] == 's' && szModel[10] == 'm') {
		// Set entity rendering (GlowShell).
		Set_Rendering(iEnt, kRenderFxGlowShell, 200, 200, 200, kRenderNormal, 16)

		// Set entity colored trail.
		message_begin(MSG_BROADCAST, SVC_TEMPENTITY)
		write_byte(TE_BEAMFOLLOW) // TE id.
		write_short(iEnt) // entity id.
		write_short(g_iTrailSpr) // sprite id.
		write_byte(10) // life time.
		write_byte(5) // width.
		write_byte(200) // r.
		write_byte(200) // g.
		write_byte(200) // b.
		write_byte(200) // brightness.
		message_end()

		// Set grenade ID 
		set_entvar(iEnt, var_flTimeStepSound, GRENADE_ID)
	}

	// Set entity (w_ model).
	if (equal(szModel, "models/w_smokegrenade.mdl")) {	
		engfunc(EngFunc_SetModel, iEnt, g_w_szFlareModel) 
		return FMRES_SUPERCEDE // Block hook from set entity model.
	}
	
	return FMRES_IGNORED
}

// Hook called after grenade think.
public GrenadeThink_Post(iEnt) {
	// Invalid entity.
	if (!is_entity(iEnt))
		return HAM_IGNORED

	if (get_entvar(iEnt, var_dmgtime) > get_gametime())
		return HAM_IGNORED

	// Grenade will explode.
	switch (get_entvar(iEnt, var_flTimeStepSound)) {
		case GRENADE_ID: {
			Make_Explode(iEnt)
			return HAM_SUPERCEDE // Block propert of smokegrenade.
		}
	}
	
	return HAM_IGNORED
}

Make_Explode(const iEnt) {
	// Get origin of the grenade. 
	static Float:flOrigin[3], iVictim, iOwner;
	get_entvar(iEnt, var_origin, flOrigin)

	// Get owner of entity.
	iOwner = get_entvar(iEnt, var_owner)

	// Invalid owner.
	if (!is_user_connected(iOwner)) {
		engfunc(EngFunc_RemoveEntity, iEnt)
		return;
	}
	
	// Owner is a zombie.
	if (ze_is_user_zombie(iOwner)) {
		engfunc(EngFunc_RemoveEntity, iEnt)
		return;
	}

	// Make the explosion
	Create_Blast3(flOrigin)

	// Play emit sound (random explode sound).
	static szSound[SOUND_MAX_LENGTH]
	ArrayGetString(g_szExplodeSound, random_num(0, ArraySize(g_szExplodeSound)-1), szSound, charsmax(szSound))
	emit_sound(iEnt, CHAN_WEAPON, szSound, VOL_NORM, ATTN_NORM, 0, PITCH_NORM)

	if (get_pcvar_num(g_pCvarExplodeSpr)) {
		// Remove playing sound.
		static bitsTeFlag;
		bitsTeFlag |= TE_EXPLFLAG_NOSOUND
		
		// Show explode sprite in position grenade.
		message_begin(MSG_BROADCAST, SVC_TEMPENTITY)
		write_byte(TE_EXPLOSION) // TE id.
		write_coord_f(flOrigin[0]) // position x.
		write_coord_f(flOrigin[1]) // y.
		write_coord_f(flOrigin[2]+64) // z. 
		write_short(g_iExplodeSpr)	// sprite id.
		write_byte(45)	// scale.
		write_byte(20)	// framerate.
		write_byte(bitsTeFlag)	// flags.
		message_end()	
	}	

	static Float:flVictimOrigin[3], Float:flDistance, Float:flFraction, Float:flDamage, iTrace
	
	if (!get_pcvar_num(g_pCvarHitType)) {
		iVictim = NULLENT;
		// Find for victim in sphere entity.
		while ((iVictim = engfunc(EngFunc_FindEntityInSphere, iVictim, flOrigin, get_pcvar_float(g_pCvarNadeRaduis)))) {
			// Not a alive. 
			if (!is_user_alive(iVictim))
				continue;

			// Get origin of victim.
			get_entvar(iVictim, var_origin, flVictimOrigin)

			// Get distance between victim and entity.
			flDistance = vector_distance(flOrigin, flVictimOrigin)
			
			// Get a damage (damage - distance).
			flDamage = get_pcvar_float(g_pCvarNadeDamage) - flDistance

			// Victim is not a zombie.
			if (!ze_is_user_zombie(iVictim))
				continue; 

			// Victim is far for entity.
			if (flDamage <= 1.0)
				continue;

			// Victim is owner of entity.
			if (iOwner == iVictim)
				continue;

			// Damage a victim.
			rg_multidmg_clear()
			rg_multidmg_add(iEnt, iVictim, flDamage, DMG_GRENADE)
			rg_multidmg_apply(iEnt, iOwner)
		}
	} else {
		// Damage a human.
		for (iVictim = 1; iVictim <= get_member_game(m_nMaxPlayers); iVictim++) {
			// Not a alive.
			if (!is_user_alive(iVictim))
				continue;
				
			// Victim is not a zombie.
			if (!ze_is_user_zombie(iVictim))
				continue;
			
			// Get origin of victim.
			get_entvar(iVictim, var_origin, flVictimOrigin)
			
			// Get distance between nade and player
			flDistance = vector_distance(flOrigin, flVictimOrigin)
			
			// Distance is less from raduis.
			if(flDistance > get_pcvar_float(g_pCvarNadeRaduis))
				continue;
			
			flOrigin[2] += 2.0
			engfunc(EngFunc_TraceLine, flOrigin, flVictimOrigin, DONT_IGNORE_MONSTERS, iEnt, iTrace)
			flOrigin[2] -= 2.0
			
			// Get trace fraction.
			get_pmtrace(iTrace, pmt_fraction, flFraction)
			
			if(flFraction != 1.0 && get_tr2(iTrace, TR_pHit) != iVictim)
				continue;
			
			// Get a damage (damage - distance).
			flDamage = get_pcvar_float(g_pCvarNadeDamage) - flDistance			
			
			// Victim is far for entity.
			if (flDamage <= 1.0)
				continue;

			// Victim is owner of entity.
			if (iOwner == iVictim)
				continue;
			
			// Damage a victim.
			rg_multidmg_clear()
			rg_multidmg_add(iEnt, iVictim, flDamage, DMG_GRENADE)
			rg_multidmg_apply(iEnt, iOwner)
		}
		
		// Free the trace handler
		free_tr2(iTrace)
	}
	
	// Remove entity
	engfunc(EngFunc_RemoveEntity, iEnt)
}

// Explode Grenade: Freeze Blast
Create_Blast3(const Float:flOrigin[3])
{
	// Smallest ring
	message_begin_f(MSG_PVS, SVC_TEMPENTITY, flOrigin)
	write_byte(TE_BEAMCYLINDER) // TE id.
	write_coord_f(flOrigin[0]) // position x.
	write_coord_f(flOrigin[1]) // y.
	write_coord_f(flOrigin[2]) // z.
	write_coord_f(flOrigin[0]) // x axis
	write_coord_f(flOrigin[1]) // y axis.
	write_coord_f(flOrigin[2]+385.0) // z axis.
	write_short(g_iRingSpr) // sprite.
	write_byte(0) // startframe.
	write_byte(0) // framerate.
	write_byte(4) // life.
	write_byte(60) // width.
	write_byte(0) // noise.
	write_byte(200) // red.
	write_byte(200) // green.
	write_byte(200) // blue.
	write_byte(200) // brightness.
	write_byte(0) // speed.
	message_end()
	
	// Medium ring
	message_begin_f(MSG_PVS, SVC_TEMPENTITY, flOrigin)
	write_byte(TE_BEAMCYLINDER) // TE id.
	write_coord_f(flOrigin[0]) // position x.
	write_coord_f(flOrigin[1]) // y.
	write_coord_f(flOrigin[2]) // z.
	write_coord_f(flOrigin[0]) // x axis.
	write_coord_f(flOrigin[1]) // y axis.
	write_coord_f(flOrigin[2]+470.0) // z axis.
	write_short(g_iRingSpr) // sprite.
	write_byte(0) // startframe.
	write_byte(0) // framerate.
	write_byte(4) // life.
	write_byte(60) // width.
	write_byte(0) // noise.
	write_byte(200) // red.
	write_byte(200) // green.
	write_byte(200) // blue.
	write_byte(200) // brightness.
	write_byte(0) // speed.
	message_end()
	
	// Largest ring
	message_begin_f(MSG_PVS, SVC_TEMPENTITY, flOrigin)
	write_byte(TE_BEAMCYLINDER) // TE id.
	write_coord_f(flOrigin[0]) // position x.
	write_coord_f(flOrigin[1]) // y.
	write_coord_f(flOrigin[2]) // z.
	write_coord_f(flOrigin[0]) // x axis.
	write_coord_f(flOrigin[1]) // y axis.
	write_coord_f(flOrigin[2]+555.0) // z axis.
	write_short(g_iRingSpr) // sprite.
	write_byte(0) // startframe.
	write_byte(0) // framerate.
	write_byte(4) // life.
	write_byte(60) // width.
	write_byte(0) // noise.
	write_byte(200) // red.
	write_byte(200) // green.
	write_byte(200) // blue.
	write_byte(200) // brightness.
	write_byte(0) // speed.
	message_end()
}